/*********************************************
 * vim:sw=2:ts=8:si
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote device and sensor
 * UDP and HTTP interface 
        url looks like this http://baseurl/password/command
        or http://baseurl/password/
 *
 * Chip type           : Atmega88 or Atmega168 with ENC28J60
 * Note: there is a version number in the text. Search for tuxgraphics
 *********************************************/
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "net.h"

// listen port for tcp/www (max range 1-254)
#define MYWWWPORT 80

#if 0
// listen port for udp
#define MYUDPPORT 1200
#endif

#define BUFFER_SIZE 450
static uint8_t buf[BUFFER_SIZE+1];

// the password string (only the first 5 char checked), (only a-z,0-9,_ characters):
static char password[] = "secret"; // must not be longer than 9 char

uint8_t
verify_password(char *str) {
  // the first characters of the received string are
  // a simple password/cookie:
  if (strncmp(password, str, 5) == 0)
    return 1;
  return 0;
}

// takes a string of the form password/commandNumber and analyse it
// return values: -1 invalid password, otherwise command number
//                -2 no command given but password valid
//                -3 valid password, no command and no trailing "/"
int8_t
analyse_get_url(char *str) {
  uint8_t i = 0;
  while (password[i]) {
    if (*str != password[i])
      return -1;
    str++;
    i++;
  }

  // is is now one char after the password
  if (*str == '/')
    str++;
  else
    return(-3);

  // check the first char, garbage after this is ignored (including a slash)
  if (*str < 0x3a && *str > 0x2f) {
    // is a ASCII number, return it
    return (*str - 0x30);
  }
  return -2;
}

// answer HTTP/1.0 301 Moved Permanently\r\nLocation: password/\r\n\r\n
// to redirect to the url ending in a slash
uint16_t
moved_perm(uint8_t *buf) {
  uint16_t plen;
  plen = fill_tcp_data_p(buf, 0,
      PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
  plen = fill_tcp_data(buf, plen, password);
  plen = fill_tcp_data_p(buf, plen,
      PSTR("/\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
  plen = fill_tcp_data_p(buf, plen,
      PSTR("<h1>301 Moved Permanently</h1>\n"));
  plen = fill_tcp_data_p(buf, plen,
      PSTR("add a trailing slash to the url\n"));
  return plen;
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t
print_webpage(uint8_t *buf, uint8_t on_off) {
  uint16_t plen;
  plen = fill_tcp_data_p(buf, 0,
    PSTR(
      "HTTP/1.0 200 OK\r\n"
      "Content-Type: text/html\r\n"
      "Pragma: no-cache\r\n"
      "\r\n"
    )
  );

  plen = fill_tcp_data_p(buf, plen, (on_off ? PSTR("ON") : PSTR("OFF")));
  plen = fill_tcp_data_p(buf, plen,
      PSTR("<a href=\".\">[refresh]</a>\n<p><a href=\"."));

  if (on_off)
    plen = fill_tcp_data_p(buf, plen, PSTR("/0\">Switch off</a><p>"));
  else
    plen = fill_tcp_data_p(buf, plen, PSTR("/1\">Switch on</a><p>"));
  return(plen);
}

uint8_t
hex2int(uint8_t c) {
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'a' && c <= 'f')
    return c - 'a' + 10;
  if (c >= 'A' && c <= 'F')
    return c - 'A' + 10;
  return 0;
}

int
main(void) {
  uint16_t plen;
  uint16_t dat_p;
  int8_t cmd;
  uint8_t baby = 0;
#if 0
  uint8_t cmd_pos = 0;
  uint8_t payloadlen = 0;
  char str[30];
  char cmdval;
#endif
  
  // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
  // full external speed)
  // set the clock prescaler. First write CLKPCE to enable setting of clock the
  // next four instructions.
  CLKPR = (1<<CLKPCE); // change enable
  CLKPR = 0; // "no pre-scaler"
  delay_ms(1);

  {
    // read config from eeprom
    uint8_t mac[6];
    uint8_t ip[4];

    mac[0] = 0x00;
    mac[1] = 0x50;
    mac[2] = 0xc2;
    mac[3] = 0x97;
    eeprom_read_block(&mac[4], (const uint8_t*)0x1f0, 2);
    if (mac[4] == 0xff) {
      mac[4] = 0x20;
      mac[5] = 0x00;
      baby = 1;
    }

//    eeprom_read_block(ip, (const uint8_t*)0x1f2, 4);
//    if (ip[0] == 0xff) {
      ip[0] = 192;
      ip[1] = 168;
      ip[2] = 1;
      ip[3] = 247;
//    }

    enc28j60Init(mac);
    enc28j60LedConfig(LED_LINK, LED_TXRX, LED_STRETCH_70MS, 1);
    init_ip_arp_udp_tcp(mac, ip, MYWWWPORT);
  }

  // LED
  DDRB |= (1<<DDB1);
  //PORTB |= (1<<PB1);

  while (1) {
    // get the next new packet:
    plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

    // plen will be non-zero if there is a valid packet (without crc error)
    if (plen == 0)
      continue;

    // ARP
    if (eth_type_is_arp_and_my_ip(buf, plen)) {
      make_arp_answer_from_request(buf);
      continue;
    }

    // check if the packet is for me.
    if (!eth_type_is_ip_and_my_ip(buf, plen))
      continue;

    // ICMP Echo
    if (buf[IP_PROTO_P] == IP_PROTO_ICMP_V) {
      if (buf[ICMP_TYPE_P] == ICMP_TYPE_ECHOREQUEST_V) {
        make_echo_reply_from_request(buf,plen);
        continue;
      }
    }

    // tcp port www start, compare only the lower byte
    if (buf[IP_PROTO_P] == IP_PROTO_TCP_V && buf[TCP_DST_PORT_H_P] == 0 && buf[TCP_DST_PORT_L_P] == MYWWWPORT) {

      // SYN
      if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V) {
        make_tcp_synack_from_syn(buf);
        // make_tcp_synack_from_syn does already send the syn,ack
        continue;
      }

      // ACK
      if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V) {
        init_len_info(buf); // init some data structures
        dat_p = get_tcp_data_pointer();
	// XXX あれ？長さは？

        if (dat_p == 0) {
          if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V) {
            // finack, answer with ack
            make_tcp_ack_from_any(buf);
          }
          // just an ack with no data, wait for next packet
          continue;
        }

        if (strncmp("GET ", (char *)&(buf[dat_p]), 4) != 0) {
          // HTTP request method is not GET.
          plen = fill_tcp_data_p(buf, 0, PSTR(
            "HTTP/1.0 200 OK\r\n"
            "Content-Type: text/html\r\n"
            "\r\n"
            "<h1>200 OK</h1>"
          ));
          goto SENDTCP;
        }

        if (strncmp("/ ",(char *)&(buf[dat_p+4]),2) == 0) {
          plen = fill_tcp_data_p(buf, 0, PSTR(
            "HTTP/1.0 200 OK\r\n"
            "Content-Type: text/html\r\n"
            "\r\n"
          ));
          plen = fill_tcp_data_p(buf, plen, PSTR(
              "<p>Usage: http://host_or_ip/password</p>\n"));
          goto SENDTCP;
        }

	if (baby) {
	  char c1 = buf[dat_p + 5];
	  char c2 = buf[dat_p + 6];
	  char c3 = buf[dat_p + 7];
	  char c4 = buf[dat_p + 8];
	  char s[3];
	  uint8_t i;

	  if (isxdigit(c1) && isxdigit(c2) && isxdigit(c3) && c4 == '/') {
	    eeprom_write_byte((uint8_t*)0x1f0, 0x20 + hex2int(c1));
	    eeprom_write_byte((uint8_t*)0x1f1, (hex2int(c2) << 4) + hex2int(c3));
	    baby = 0;
	  }

	  plen = 0;
	  plen = fill_tcp_data(buf, plen, "0050c297");

	  for (i = 0; i < 2; ++i) {
	    c1 = eeprom_read_byte((const uint8_t*)(0x1f0 + i));
	    c2 = (c1 >> 4) & 0x0f;
	    c3 = c1 & 0x0f;
	    s[0] = c2 < 10 ? ('0' + c2) : ('a' - 10 + c2);
	    s[1] = c3 < 10 ? ('0' + c3) : ('a' - 10 + c3);
	    s[2] = '\0';
	    plen = fill_tcp_data(buf, plen, s);
	  }
          goto SENDTCP;
	}

        cmd = analyse_get_url((char *)&(buf[dat_p+5]));
        switch (cmd) {
        case -3:
          plen = moved_perm(buf);
          goto SENDTCP;

	case -2:
	  break;

        case -1:
          plen = fill_tcp_data_p(buf, 0, PSTR(
            "HTTP/1.0 401 Unauthorized\r\n"
            "Content-Type: text/html\r\n"
            "\r\n"
            "<h1>401 Unauthorized</h1>"
          ));
          goto SENDTCP;

        case 1:
	  PORTB &= ~(1<<PB1);
          break;

        case 0:
	  PORTB |= (1<<PB1);
          break;
        }
	plen = print_webpage(buf, !(PORTB & (1<<PB1)));

SENDTCP:
        make_tcp_ack_from_any(buf); // send ack for http get
        make_tcp_ack_with_data(buf, plen); // send data
        continue;
      }
    }

#if 0
    // udp start, we listen on udp port 1200=0x4B0
    if (buf[IP_PROTO_P] == IP_PROTO_UDP_V && buf[UDP_DST_PORT_H_P] == 4 && buf[UDP_DST_PORT_L_P] == 0xb0) {
      payloadlen = buf[UDP_LEN_L_P] - UDP_HEADER_LEN;

      if (!verify_password((char *)&(buf[UDP_DATA_P]))) {
        // Invalid password
        strcpy(str, "e=invalid_pw");
        make_udp_reply_from_request(buf, str, strlen(str), MYUDPPORT);
        continue;
      }

      // find the first comma which indicates 
      // the start of a command:
      cmd_pos = 0;
      while (cmd_pos < payloadlen) {
        ++cmd_pos;
        if (buf[UDP_DATA_P+cmd_pos] == ',') {
          ++cmd_pos; // put on start of cmd
          break;
        }
      }

      // a command is one char and a value. At
      // least 3 characters long. It has an '=' on
      // position 2:
      if (cmd_pos < 2 || cmd_pos > payloadlen-3 || buf[UDP_DATA_P+cmd_pos+1] != '=') {
        strcpy(str, "e=no_cmd");
        make_udp_reply_from_request(buf, str, strlen(str), MYUDPPORT);
        continue;
      }

      // supported commands are
      // t=1 t=0 t=?
      if (buf[UDP_DATA_P+cmd_pos] == 't') {
        cmdval = buf[UDP_DATA_P+cmd_pos+2];
        if (cmdval == '1') {
          PORTB |= (1<<PB1);// transistor on
          strcpy(str, "t=1");
        } else
        if (cmdval == '0') {
          PORTB &= ~(1<<PB1);// transistor off
          strcpy(str, "t=0");
        } else
        if (cmdval == '?') {
          if (PORTB & (1<<PB1))
            strcpy(str, "t=1");
          else
            strcpy(str, "t=0");
        } else
          strcpy(str,"e=no_such_cmd");
      }
      make_udp_reply_from_request(buf,str,strlen(str),MYUDPPORT);
    }
#endif

  }
  return 0;
}
